Unit DotList;

Interface

Uses Windows,SysUtils,Classes,DotDef;

Type

{ TDotList class }

  TDotList = class(TPersistent)
  private
    FList: PPointerList;
    FCount: Integer;
    FCapacity: Integer;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
  protected
    function  GetStyle(Index: Integer): PDotStyle;
    function  GetStyleName(Index: Integer):String;
    procedure Grow; virtual;
    procedure PutStyle(Index: Integer; Item: PDotStyle);
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetCount(NewCount: Integer);
    procedure Changed; dynamic;
    procedure Changing; dynamic;
    procedure AssignTo(Dest: TPersistent); override;
  public
    procedure Assign(Source: TPersistent); override;
    constructor Create;
    destructor Destroy; override;
    function AddStyle(Item: PDotStyle): Integer;
    procedure Clear;
    class procedure Error(const Msg: string; Data: Integer); virtual;
    function Expand: TDotList;
    function IndexOf(Item: PDotStyle):Integer;
    function IndexOfName(Name:String):Integer;
    procedure SaveToFile(FileName:String);
    procedure LoadFromFile(FileName:String);
    property Capacity: Integer read FCapacity write SetCapacity;
    property Count: Integer read FCount write SetCount;
    property Items[Index: Integer]:PDotStyle read GetStyle write PutStyle; default;
    property Names[Index: Integer]:String read GetStyleName;
    property List: PPointerList read FList;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
  end;

Implementation

Uses Consts;

{ TDotList }

function  TDotList.GetStyleName(Index: Integer):String;
 begin
   if (Index<0) or (Index>=FCount) then Exit;
   Result:=GetStyle(Index).FStyleName;
 end;

procedure TDotList.AssignTo(Dest: TPersistent);
 begin
  if Dest is TDotList then Dest.Assign(self)
  else inherited AssignTo(Dest);
 end;

procedure TDotList.Assign(Source: TPersistent);
Var i,c:Integer;
    p:PDotStyle;
 begin
   inherited Assign(Source);
   if (Source is TDotList) then
    begin
      Clear;
      c:=(Source as TDotList).Count;
      for i:=0 to (c-1) do
       begin
        p:=PDotStyle(AllocMem(sizeOf(TDotStyle)));
        p^:=(Source as TDotList).Items[i]^;
        AddStyle(p);
       end;
       p:=nil;
    end;
 end;

constructor TDotList.Create;
begin
  inherited Create;
  FOnChange:=nil;
  FOnChanging:=nil;
end;

destructor TDotList.Destroy;
begin
  FOnChange:=nil;
  FOnChanging:=nil;
  Clear;
end;

function TDotList.AddStyle(Item: PDotStyle): Integer;
begin
  Changing;
  Result := FCount; { FStyleID}
  if Result = FCapacity then Grow;
  { Setting StyleID !!! }
  Item^.FStyleID:=FCount;
  FList^[Result] := Pointer(Item);
  Inc(FCount);
  Changed;
end;

procedure TDotList.Clear;
Var i:Integer;
begin
  for i:=0 to (FCount-1) do
     FreeMem(PDotStyle(FList^[i]));
  SetCount(0);
  SetCapacity(0);
end;

class procedure TDotList.Error(const Msg: string; Data: Integer);

  function ReturnAddr: Pointer;
  asm
          MOV     EAX,[EBP+4]
  end;

begin
  raise EListError.CreateFmt(Msg, [Data]) at ReturnAddr;
end;

function TDotList.Expand: TDotList;
begin
  if FCount = FCapacity then Grow;
  Result := Self;
end;


function TDotList.GetStyle(Index: Integer): PDotStyle;
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Result := FList^[Index];
end;

procedure TDotList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TDotList.IndexOf(Item: PDotStyle): Integer;
begin
  Result := 0;
  while (Result < FCount) and ((PDotStyle(FList^[Result])) <> Item) do Inc(Result);
  if Result = FCount then Result := -1;
end;

procedure TDotList.PutStyle(Index: Integer; Item: PDotStyle);
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  if (FList^[Index]<>nil) then FreeMem(PDotStyle(FList^[Index]));
  FList^[Index] := Pointer(Item);
end;

procedure TDotList.SetCapacity(NewCapacity: Integer);
begin
  if (NewCapacity < FCount) or (NewCapacity > MaxListSize) then
    Error(SListCapacityError, NewCapacity);
  if NewCapacity <> FCapacity then
  begin
    ReallocMem(FList, NewCapacity * SizeOf(Pointer));
    FCapacity := NewCapacity;
  end;
end;

procedure TDotList.SetCount(NewCount: Integer);
begin
  if (NewCount < 0) or (NewCount > MaxListSize) then
    Error(SListCountError, NewCount);
  if NewCount > FCapacity then SetCapacity(NewCount);
  if NewCount > FCount then
    FillChar(FList^[FCount], (NewCount - FCount) * SizeOf(Pointer), 0);
  FCount := NewCount;
end;

procedure TDotList.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDotList.Changing;
begin
  if Assigned(FOnChanging) then FOnChanging(Self);
end;

procedure TDotList.SaveToFile(FileName:String);
Var F:File of TDotStyle;
    i:Integer;
 begin
  {}
   AssignFile(F,FileName);
   Rewrite(F);
   for i:=0 to (FCount-1) do
    begin
      Write(F,(PDotStyle(FList[i]))^);
    end;
   CloseFile(F);
 end;

function TDOTList.IndexOfName(Name:String):Integer;
 begin
  Result := 0;
  while (Result < FCount) and ((PDotStyle(FList^[Result])^.FStyleName) <> Name) do Inc(Result);
  if Result = FCount then Result := -1;

 end;

procedure TDotList.LoadFromFile(FileName:String);
Var F:File of TDotStyle;
    P:PDotStyle;
    T:TDotStyle;
 begin
  if not(FileExists(FileName)) then Exit;
  Clear;
  AssignFile(F,FileName);
  Reset(F);
   while (not(EOF(F))) do
     begin
       Read(F,T);
       P:=PDotStyle(AllocMem(SizeOf(TDotStyle)));
       P^:=T;
       AddStyle(P);
     end;
  CloseFile(F);
  P:=nil;
 end;

end.