unit Rulers;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, ExtCtrls;

{Copyright (c) 1996 DinoByte Software  All Rights Reserved}
{3D Ruler component for Delphi 16 & 32 bit compilers.  This ruler can be
 displayed in a horizontal or vertical direction.  It has properties which
 allow automatic displaying of alignment indicators and tab markers.}

{A note about variable, procedure, and function names:
 The code may be easier to follow if you understand my naming conventions.
 They may seem a little strange but they work well for me.  All variables
 begin with letters as follows:
    Boolean - flg     ex. flgReady
    Byte - b          ex. bCnt
    Integer - i       ex. iLeft
    Word - w          ex. wSegment
    LongInt - l       ex. lTag
    Strings - s       ex. sName
    Objects - a       ex. aFont
 Field variables are an exception to this; they begin with the letter F.
 In addition to this I usually name the arguments of property procedures
 with their Type letter and the word 'Value'.  For example, an Integer
 argument would be named iValue.  This is useful when using Delphi's
 'View Watches' window. I don't need as many variables listed this way.

  Any internal procedures or functions begin with a lower case letter.
  This makes it easy to tell where the source can be found.  Not very
  important in a file this size but very handy for large projects.
  Generic routines are listed first, followed by the Constructor,
  Destructor and any of the control's events.  Next are the 'Get' functions
  and then the 'Set' procedures. }

const
  VERSION_ID = 101; {vs 1.01}
  MAXSEGMENTS = 64;
  MAXTABS = 30;

type
  TDirection = (dirHorz, dirVert);

  TSegStyle = (ssDefault, ssEighth, ssFifth,
   ssQuarter, ssSixteenth, ssTenth, ssThird);

  TTabPosition = (tpBottom, tpLeft, tpRight, tpTop);

  TIntArray = array[0..MAXTABS-1] of Integer;
  TRealArray = array[0..MAXTABS-1] of Real;

  TRuler = class(TGraphicControl)
  private
    FBorderStyle: TBorderStyle;
    FCornerRect: TRect;
    FCornerSize: Integer;
    FCornerStyle: TBevelStyle;
    FCtl3D: Boolean;
    FDelphi: Boolean;
    FDirection: TDirection;
    FMarkColor: TColor;
    FMarkerPos: Integer;
    FMax, FMin: Integer;
    FNumLeft: Integer;
    FNumTop: Integer;
    FOffset: Integer;
    FOldX1, FOldX2, FOldY1, FOldY2: Integer;
    FPixelsPerSegment: Integer;
    FRulerAlign: TAlign;
    FSegColor: TColor;
    FSegCount: Byte;
    FSegments: array[1..MAXSEGMENTS] of Byte;
    FSegSize: Byte;
    FSegStyle: TSegStyle;
    FTabCount: Integer;
    FTabs: array[0..MAXTABS] of Integer;
    FTabsEnabled: Boolean;
    FTabGlyph: TBitmap;
    FTabPosition: TTabPosition;
    FVersion: Integer;
    function getFont: TFont;
    function getSegments(bIndex: Byte): Byte;
    function getTabs(iIndex: Integer): Integer;
    procedure drawBorders(var aRect: TRect);
    procedure drawTabs(aRect: TRect);
    procedure fixSizes(var aRect: TRect;
               iWidth, iHeight: Integer; var wSeg, wMax: Word);
    procedure glyphChange(Sender: TObject);
    procedure resetSegInfo;
    procedure setBorderStyle(aValue: TBorderStyle);
    procedure setCornerSize(iValue: Integer);
    procedure setCornerStyle(aValue: TBevelStyle);
    procedure setCtl3D(fValue: Boolean);
    procedure setDirection(aValue: TDirection);
    procedure setFont(aFont: TFont);
    procedure setMarkerPos(iValue: Integer);
    procedure setNumLeft(iValue: Integer);
    procedure setNumTop(iValue: Integer);
    procedure setOffset(iValue: Integer);
    procedure setRulerAlign(Value: TAlign);
    procedure setSegColor(aValue: TColor);
    procedure setSegCount(bValue: Byte);
    procedure setSegments(bIndex, bValue: Byte);
    procedure setSegSize(bValue: Byte);
    procedure setSegStyle(aValue: TSegStyle);
    procedure setTabGlyph(aValue: TBitmap);
    procedure setTabPosition(aValue: TTabPosition);
    procedure setTabs(iIndex, iValue: Integer);
    procedure setTabsEnabled(flgValue: Boolean);
    procedure setVersion(iValue: Integer);
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AddTabByInch(rValue: Real): Integer;
    function AddTabByPixel(iValue: Integer): Integer;
    function AddTabBySegment(iValue: Integer): Integer;
    function DeleteTabs(iFirst, iCount: Integer): Integer;
    function GetActualPos(iValue: Integer): Integer;
    function GetRulerPos(iValue: Integer): Integer;
    function GetTabByInch(iIndex: Integer): Real;
    function GetTabByPixel(iIndex: Integer): Integer;
    function GetTabBySegment(iIndex: Integer): Integer;
    function PointOnTab(iX, iY: Integer): Integer;
    function SetTabsByInch(iFirst, iCount: Integer;
              var rValues: TRealArray): Integer;
    function SetTabsByPixel(iFirst, iCount: Integer;
              var iValues: TIntArray): Integer;
    function SetTabsBySegment(iFirst, iCount: Integer;
              var iValues: TIntArray): Integer;
    procedure ResetTabs;
    property MarkerPos: Integer
              read FMarkerPos write setMarkerPos default -1;
    property PixelsPerSegment: Integer read FPixelsPerSegment;
    property SegmentCount: Byte read FSegCount write setSegCount default 12;
    property Segments[i: Byte]: Byte read getSegments write setSegments;
    property TabCount: Integer read FTabCount;
    property Tabs[i: Integer]: Integer read getTabs write setTabs;
  published
    property AlignRuler: TAlign
     read FRulerAlign write setRulerAlign default alNone;
    property BorderStyle: TBorderStyle
              read FBorderStyle write setBorderStyle default bsNone;
    property Color default clBtnFace;
    property CornerSize: Integer
              read FCornerSize write setCornerSize default 0;
    property CornerStyle: TBevelStyle
              read FCornerStyle write setCornerStyle default bsRaised;
    property Ctl3D: Boolean read FCtl3D write setCtl3D default True;
    property Direction: TDirection
              read FDirection write setDirection default dirHorz;
    property DragCursor;
    property DragMode;
    property Font: TFont read getFont write setFont;
    property Height default 33;
    property MarkerColor: TColor
              read FMarkColor write FMarkColor default clBtnHighlight;
    property NumeralLeft: Integer read FNumLeft write setNumLeft default 2;
    property NumeralTop: Integer read FNumTop write setNumTop default 12;
    property Offset: Integer
              read FOffset write setOffset default 0;
    property ParentColor default True;
    property ParentFont default False;
    property Visible;
    property Width default 768;
    property SegmentColor: TColor
              read FSegColor write setSegColor default clBlack;
    property SegmentSize: Byte read FSegSize write setSegSize default 6;
    property SegmentStyle: TSegStyle
              read FSegStyle write setSegStyle default ssDefault;
    property TabsEnabled: Boolean
              read FTabsEnabled write setTabsEnabled default False;
    property TabGlyph: TBitmap
              read FTabGlyph write setTabGlyph;
    property TabPosition: TTabPosition
              read FTabPosition write setTabPosition default tpTop;
    property Version: Integer read FVersion write setVersion default VERSION_ID;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

procedure Register;

implementation

{$IFDEF WINDOWS}  {Used only in 16 bit version}
  uses ToolHelp;
{$ENDIF}

{Change this to create a new default TabGlyph.}
{$R rulertab.res}

{Taken from a library of miscellaneous routines I use often.
 It decreases the size of a TRect by subtracting from each side.}
procedure dbImplodeRect(var Rect: TRect; iAdd: Integer);
begin
  Inc(Rect.Left, iAdd);
  Inc(Rect.Top, iAdd);
  Dec(Rect.Right, iAdd);
  Dec(Rect.Bottom, iAdd);
end;

{These routines, 16 and 32 bit, return True if Delphi is running}
{Thanks to Michael Bolli-Starke CIS 100566,576 for the 16 bit code.}
{$IFDEF WIN32}
function dbDelphiRunning: Boolean; {32 bit version}
var
	iHnd: Integer;
	cClassName: array[0..20] of Char;
begin
  StrPCopy(cClassName, 'TAppBuilder');
	iHnd := FindWindow(cClassName, nil);
  Result := (iHnd <> 0);
end;
{$ELSE}
function dbDelphiRunning: Boolean; {16 bit version}
const
	sDelphi: array[1..4] of string = ('DELPHI', 'W8LOSS', 'DELPHIED', 'DELPHIKB');
var
	aModule: TModuleEntry;
	iModCnt, iCnt: integer;
begin
  aModule.dwSize := Sizeof(aModule);
  iModCnt := 0;
  if ModuleFirst(@aModule) then begin
    repeat
      for iCnt := 1 to 4 do begin
        if sDelphi[iCnt] = StrPas(aModule.szModule) then Inc(iModCnt);
      end;
    until (not ModuleNext(@aModule));
  end;
  if iModCnt < 4 then Result := False
  else Result := True;
end;
{$ENDIF}

{Returns value if negative, 0 otherwise}
function dbNegative(lValue: LongInt): LongInt;
begin
  if lValue < 0 then Result := lValue
  else Result := 0;
end;

{Returns value if positive, 0 otherwise}
function dbPositive(lValue: LongInt): LongInt;
begin
  if lValue > 0 then Result := lValue
  else Result := 0;
end;

{Draws the Ruler's borders for the Paint procedure.}
procedure TRuler.drawBorders(var aRect: TRect);
begin
  with Canvas do begin
    if BorderStyle = bsSingle then begin
      Pen.Color := clBlack;
      with aRect do Rectangle(Left, Top, Right, Bottom);
      with FCornerRect do Rectangle(Left, Top, Right, Bottom);
      dbImplodeRect(aRect, 1);
      dbImplodeRect(FCornerRect, 1);
    end;
    if Ctl3D then begin
      Frame3D(Canvas, aRect, clBtnHighlight, clBtnShadow, 1);
      if CornerStyle = bsRaised then
        Frame3D(Canvas, FCornerRect, clBtnHighlight, clBtnShadow, 1)
      else
        Frame3D(Canvas, FCornerRect, clBtnShadow, clBtnHighlight, 1);
    end;
  end;
end;

{Draws the Tabs for the Paint procedure.  Just to break things up a little.}
{This is the only routine where I actually separate function for Horz and Vert}
procedure TRuler.drawTabs(aRect: TRect);
var
  iCnt, iCorner, iCut, iHalf, iHeight, iWidth: Integer;
  aColor: TColor;
  aRuleRect, aTabRect: TRect;
begin
  if TabsEnabled and (TabGlyph.Empty = False) then begin
    Canvas.Brush.Color := Color;
    aColor := TabGlyph.TransparentColor;
    iCorner := dbPositive(CornerSize);
    iWidth := TabGlyph.Width; iHeight := TabGlyph.Height;
    aTabRect := Rect(0, 0, iWidth, iHeight);
    if Direction = dirHorz then begin
      if TabPosition = tpTop then begin
        aRuleRect.Top := aRect.Top;
        aRuleRect.Bottom := aRect.Top + iHeight;
      end else begin {TabPosition = tpBottom}
        aRuleRect.Bottom := aRect.Bottom;
        aRuleRect.Top := aRect.Bottom - iHeight;
      end;
      iHalf := iWidth div 2;
      for iCnt := 0 to TabCount-1 do begin
        aRuleRect.Left := Tabs[iCnt] + iCorner - Offset - iHalf;
        aRuleRect.Right := aRuleRect.Left + iWidth;
        if aRuleRect.Right <= aRect.Left then continue;
        iCut := dbPositive(aRect.Left - aRuleRect.Left);
        Inc(aRuleRect.Left, iCut);
        Inc(aTabRect.Left, iCut);

        if aRuleRect.Left >= aRect.Right then continue;
        iCut := dbNegative(aRect.Right - aRuleRect.Right);
        Inc(aRuleRect.Right, iCut);
        Inc(aTabRect.Right, iCut);

        Canvas.BrushCopy(aRuleRect, TabGlyph, aTabRect, aColor);
        aTabRect.Left := 0;
      end;
    end else begin  {Direction = dirVert}
      if TabPosition = tpLeft then begin
        aRuleRect.Left := aRect.Left;
        aRuleRect.Right := aRect.Left + iWidth;
      end else begin {TabPosition = tpRight}
        aRuleRect.Right := aRect.Right;
        aRuleRect.Left := aRect.Right - iWidth;
      end;
      iHalf := iHeight div 2;
      for iCnt := 0 to TabCount-1 do begin
        aRuleRect.Top := Tabs[iCnt] + iCorner - Offset - iHalf;
        aRuleRect.Bottom := aRuleRect.Top + iHeight;
        if aRuleRect.Bottom <= aRect.Top then continue;
        iCut := dbPositive(aRect.Top - aRuleRect.Top);
        Inc(aRuleRect.Top, iCut);
        Inc(aTabRect.Top, iCut);

        if aRuleRect.Top >= aRect.Bottom then continue;
        iCut := dbNegative(aRect.Bottom - aRuleRect.Bottom);
        Inc(aRuleRect.Bottom, iCut);
        Inc(aTabRect.Bottom, iCut);

        Canvas.BrushCopy(aRuleRect, TabGlyph, aTabRect, aColor);
        aTabRect.Top := 0;
      end;
    end;
  end;
end;

{Adjusts some sizing variables for the Paint procedure.}
procedure TRuler.fixSizes(var aRect: TRect;
 iWidth, iHeight: Integer; var wSeg, wMax: Word);
begin
  if Direction = dirHorz then begin
    if CornerSize > 0 then begin
      Inc(aRect.Left, wSeg);
      wMax := iWidth;
      FMin := wSeg; FMax := iWidth;
      FCornerRect := Rect(0, 0, aRect.Left, aRect.Bottom);
    end else begin
      Dec(aRect.Right, wSeg);
      wMax := iWidth - wSeg;
      FMin := 0; FMax := iWidth - wSeg;
      FCornerRect := Rect(aRect.Right, 0, aRect.Right + wSeg, aRect.Bottom);
      wSeg := 0;
    end;
  end else begin
    if CornerSize > 0 then begin
      Inc(aRect.Top, wSeg);
      wMax := iHeight;
      FMin := wSeg; FMax := iHeight;
      FCornerRect := Rect(0, 0, aRect.Right, aRect.Top);
    end else begin
      Dec(aRect.Bottom, wSeg);
      wMax := iHeight - wSeg;
      FMin := 0; FMax := iHeight-wSeg;
      FCornerRect := Rect(0, aRect.Bottom, aRect.Right, aRect.Bottom + wSeg);
      wSeg := 0;
    end;
  end;
end;

{This routine simply causes the ruler to repaint if the glyph changes.}
procedure TRuler.glyphChange(Sender: TObject);
begin
  Invalidate;
end;

{The FSegments[] array controls the length of the little hashmarks which
 mark the various segments of the ruler.  FSegments[1] would be the length
 of the 1 inch mark, 2 would be 1/2 inch, 3 would be 1/3, and so on.  The
 values in the array actually get multiplied by the .SegmentLength property.}
procedure TRuler.resetSegInfo;
var
  bCnt: Byte;
begin
  for bCnt := 1 to 64 do FSegments[bCnt] := 0;
end;

{********************************* TRuler **********************************}

constructor TRuler.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDelphi := dbDelphiRunning;
  AlignRuler := alNone;
  BorderStyle := bsNone;
  CornerSize := 0;
  CornerStyle := bsRaised;
  Ctl3D := True;
  FDirection := dirHorz;
  FOldX1 := -1; FOldX2 := -1;
  FOldY1 := -1; FOldY2 := -1;
  Height := 33;
  MarkerColor := clBtnHighlight;
  ParentColor := True;
  ParentFont := False;
  SegmentColor := clBlack;
  SegmentCount := 12;
  Offset := 0;
  SegmentSize := 6;
  SegmentStyle := ssDefault;
  Width := 768;
  TabsEnabled := False;
  TabPosition := tpTop;
  FTabGlyph := TBitmap.Create;
  FTabGlyph.Handle := LoadBitmap(hInstance, 'TAB_IMAGE');
  FTabGlyph.OnChange := glyphChange;
  FVersion := VERSION_ID;
end;

destructor TRuler.Destroy;
begin
  if Assigned(FTabGlyph) then FTabGlyph.Free;
  inherited Destroy;
end;

procedure TRuler.Paint;
var
  bCnt: Byte;
  iCnt, iPos: Integer;
  wFrac, wHash, wMax, wSeg: word;
  FCornerRect, aRect: TRect;
begin
  {Prevent any display if Delphi is not running!}
  {if FDelphi = False then exit;}

  {Reset Marker Position off of control}
  FOldX1 := -1; FOldX2 := -1;
  FOldY1 := -1; FOldY2 := -1;
  {Get the number of pixels in each segment.
   Note that this only works properly if (PixelsPerInch MOD SegmentCount = 0}
  wFrac := Screen.PixelsPerInch div SegmentCount;
  wHash := 0;
  wSeg := Abs(CornerSize);
  Canvas.Font.Color := Font.Color;
  with Canvas do begin
    Pen.Mode := pmCopy;
    Brush.Color := Color;
    Brush.Style := bsSolid;
    aRect := ClientRect;
    FillRect(aRect);
    fixSizes(aRect, Width, Height, wSeg, wMax);
    drawBorders(aRect);
    Brush.Style := bsClear;
    Pen.Color := SegmentColor;
    while wSeg < wMax + Offset do begin
      if wHash > 0 then begin
        for bCnt := SegmentCount downto 1 do begin
          if FSegments[bCnt] > 0 then begin
            if (wHash mod bCnt = 0) then begin
              iPos := wSeg - Offset;
              if iPos > CornerSize then begin
                if Direction = dirHorz then begin
                  MoveTo(iPos, 1);
                  LineTo(iPos, FSegments[bCnt]);
                end else begin
                  MoveTo(1, iPos);
                  LineTo(FSegments[bCnt], iPos);
                end;
              end;
              break;
            end;
          end;
        end;
      end;
      if (wHash > 0) and (wHash mod SegmentCount = 0) then begin
        if Direction = dirHorz then
          TextRect(aRect,
           PenPos.X+FNumLeft, FNumTop, IntToStr(wHash div SegmentCount))
        else
          TextRect(aRect,
           FNumLeft, PenPos.Y+FNumTop, IntToStr(wHash div SegmentCount));
      end;
      wHash := wHash + 1;
      wSeg := wSeg + wFrac;
    end;
    if TabsEnabled and (TabGlyph.Empty = False) then drawTabs(aRect);
  end;
end;

function TRuler.getFont: TFont;
begin
  Result := Canvas.Font;
end;

function TRuler.getSegments(bIndex: Byte): Byte;
begin
  Result := 0;
  if (bIndex > 0) and (bIndex <= SegmentCount) then Result := FSegments[bIndex];
end;

function TRuler.getTabs(iIndex: Integer): Integer;
begin
  Result := -1;
  if iIndex <= FTabCount then Result := FTabs[iIndex];
end;

procedure TRuler.setBorderStyle(aValue: TBorderStyle);
begin
  if FBorderStyle <> aValue then begin
    FBorderStyle := aValue;
    Invalidate;
  end;
end;

procedure TRuler.setCornerSize(iValue: Integer);
begin
  if FCornerSize <> iValue then begin
    FCornerSize := iValue;
    Invalidate;
  end;
end;

procedure TRuler.setCornerStyle(aValue: TBevelStyle);
begin
  if FCornerStyle <> aValue then begin
    FCornerStyle := aValue;
    Invalidate;
  end;
end;

procedure TRuler.setCtl3D(fValue: Boolean);
begin
  if fValue <> FCtl3D then begin
    FCtl3D := fValue;
    Invalidate;
  end;
end;

procedure TRuler.setDirection(aValue: TDirection);
begin
  if aValue = dirHorz then begin
    if (Align <> alLeft) and (Align <> alRight) then FDirection := aValue;
  end else begin
    if (Align <> alBottom) and (Align <> alTop) then FDirection := aValue;
  end;
  TabPosition := TabPosition;
end;

procedure TRuler.setFont(aFont: TFont);
begin
  Canvas.Font := aFont;
  Invalidate;
end;

procedure TRuler.setMarkerPos(iValue: Integer);
begin
  if (iValue >= FMin - CornerSize) and (iValue <= FMax) then begin
    Canvas.Pen.Color := MarkerColor;
    Canvas.Pen.Mode := pmXor;
    Canvas.MoveTo(FOldX1, FOldY1);
    Canvas.LineTo(FOldX2, FOldY2);
    FMarkerPos := iValue;
    if Direction = dirHorz then begin
      FOldX1 := MarkerPos + dbPositive(CornerSize);
      FOldX2 := FOldX1;
      FOldY1 := 2;
      FOldY2 := Height-2;
    end else begin
      FOldY1 := MarkerPos + dbPositive(CornerSize);
      FOldY2 := FOldY1;
      FOldX1 := 2;
      FOldX2 := Width-2;
    end;
    Canvas.MoveTo(FOldX1, FOldY1);
    Canvas.LineTo(FOldX2, FOldY2);
  end;
end;

procedure TRuler.setNumLeft(iValue: Integer);
begin
  if FNumLeft <> iValue then begin
    FNumLeft := iValue;
    Invalidate;
  end;
end;

procedure TRuler.setNumTop(iValue: Integer);
begin
  if FNumTop <> iValue then begin
    FNumTop := iValue;
    Invalidate;
  end;
end;

procedure TRuler.setOffset(iValue: Integer);
begin
  if FOffset <> iValue then begin
    FOffset := iValue;
    Invalidate;
  end;
end;

procedure TRuler.setRulerAlign(Value: TAlign);
begin
  if Value in [alTop, alBottom] then FDirection := dirHorz
  else if Value in [alLeft, alRight] then FDirection := dirVert;
  FRulerAlign := Value;
  Align := Value;
end;

procedure TRuler.setSegColor(aValue: TColor);
begin
  if FSegColor <> aValue then begin
    FSegColor := aValue;
    Invalidate;
  end;
end;

procedure TRuler.setSegCount(bValue: Byte);
begin
  if (bValue > 0) and (bValue < 65) then begin
    FSegCount := bValue;
    FPixelsPerSegment := Screen.PixelsPerInch div bValue;
    Invalidate;
  end;
end;

procedure TRuler.setSegments(bIndex, bValue: Byte);
begin
  if (bIndex > 0) and (bIndex <= SegmentCount) then begin
    if (SegmentCount mod bIndex = 0) then FSegments[bIndex] := bValue
    else FSegments[bIndex] := 0;
  end;
end;

procedure TRuler.setSegSize(bValue: Byte);
begin
  FSegSize := bValue;
  SegmentStyle := SegmentStyle;
end;

procedure TRuler.setSegStyle(aValue: TSegStyle);
begin
  resetSegInfo;
  FSegStyle := aValue;
  case aValue of
    ssDefault: begin
      FSegCount := 12;
      Segments[1] := SegmentSize;
      Segments[3] := SegmentSize * 2;
      Segments[6] := SegmentSize * 3;
      Segments[12] := SegmentSize * 4;
    end;
    ssEighth: begin
      FSegCount := 8;
      Segments[1] := SegmentSize;
      Segments[2] := SegmentSize * 2;
      Segments[4] := SegmentSize * 3;
      Segments[8] := SegmentSize * 4;
    end;
    ssFifth: begin
      FSegCount := 5;
      Segments[1] := SegmentSize;
      Segments[5] := SegmentSize * 4;
    end;
    ssQuarter: begin
      FSegCount := 4;
      Segments[1] := SegmentSize;
      Segments[2] := SegmentSize * 2;
      Segments[4] := SegmentSize * 4;
    end;
    ssSixteenth: begin
      FSegCount := 16;
      Segments[1] := SegmentSize;
      Segments[2] := SegmentSize * 2;
      Segments[4] := SegmentSize * 3;
      Segments[8] := SegmentSize * 4;
      Segments[16] := SegmentSize * 6;
    end;
    ssTenth: begin
      FSegCount := 10;
      Segments[1] := SegmentSize;
      Segments[2] := SegmentSize * 2;
      Segments[5] := SegmentSize * 3;
      Segments[10] := SegmentSize * 4;
    end;
    ssThird: begin
      FSegCount := 3;
      Segments[1] := SegmentSize;
      Segments[3] := SegmentSize * 3;
    end;
  end;
  Invalidate;
end;

procedure TRuler.setTabGlyph(aValue: TBitmap);
begin
  FTabGlyph.Assign(aValue);
  Invalidate;
end;

procedure TRuler.setTabPosition(aValue: TTabPosition);
begin
  if Direction = dirHorz then begin
    if aValue = tpLeft then aValue := tpTop
    else if aValue = tpRight then aValue := tpBottom;
  end else begin
    if aValue = tpTop then aValue := tpLeft
    else if aValue = tpBottom then aValue := tpRight;
  end;
  FTabPosition := aValue;
  Invalidate;
end;

procedure TRuler.setTabs(iIndex, iValue: Integer);
begin
  if iIndex >= MAXTABS then exit;
  if iIndex > FTabCount then FTabCount := iIndex;
  FTabs[iIndex] := iValue;
  Invalidate;
end;

procedure TRuler.setTabsEnabled(flgValue: Boolean);
begin
  if FTabsEnabled <> flgValue then begin
    FTabsEnabled := flgValue;
    Invalidate;
  end;
end;

procedure TRuler.setVersion(iValue: Integer);
begin
  {This routine does nothing.  The property can not be changed. Making it
  read-only doen't work because then it won't show up in the property editor.}
  FVersion := VERSION_ID;
end;

{******************************** Methods ***********************************
* AddTabByInch() - Adds a new tab at a specific inch location.              *
* AddTabByPixel() - Adds a new tab at a specific pixel location.            *
* AddTabBySegment() - Adds a new tab at a specific segment location.        *
* DeleteTabs() - Deletes a number of tabs from the list.                    *
* GetActualPos() - Returns an actual position on the component's canvas.    *
* GetRulerPos() - Returns the position corrected for Offset and CornerSize. *
* GetTabByInch() - Returns the position of a tab in inches.                 *
* GetTabByPixel() - Returns the position of a tab in pixels.                *
* GetTabBySegment() - Returns the position of a tab in segments.            *
* PointOnTab() - Returns the index of the tab at a X, Y position; or -1.    *
* ResetTabs - Resets all Tabs to -1 and FTabCount to 0                      *
* SetTabsByInch() - Adds or sets a number of tabs by inch locations.        *
* SetTabsByPixel() - Adds or sets a number of tabs by pixel locations.      *
* SetTabsBySegment() - Adds or sets a number of tabs by segment locations.  *
****************************************************************************}

function TRuler.AddTabByInch(rValue: Real): Integer;
var
  iValue: Integer;
begin
  if FTabCount < MAXTABS then begin
    iValue := Trunc(Screen.PixelsPerInch * rValue);
    FTabs[FTabCount] := iValue;
    Inc(FTabCount);
    Invalidate;
  end;
  Result := TabCount;
end;

function TRuler.AddTabByPixel(iValue: Integer): Integer;
begin
  if FTabCount < MAXTABS then begin
    FTabs[FTabCount] := iValue;
    Inc(FTabCount);
    Invalidate;
  end;
  Result := TabCount;
end;

function TRuler.AddTabBySegment(iValue: Integer): Integer;
begin
  if FTabCount < MAXTABS then begin
    iValue := (Screen.PixelsPerInch div SegmentCount) * iValue;
    FTabs[FTabCount] := iValue;
    Inc(FTabCount);
    Invalidate;
  end;
  Result := TabCount;
end;

function TRuler.DeleteTabs(iFirst, iCount: Integer): Integer;
var
  iCnt: Integer;
begin
  if (iFirst >= 0) and (iFirst < TabCount) then begin
    if iFirst + iCount > TabCount then iCount := TabCount - iFirst;
    for iCnt := iFirst to TabCount+iCount-1 do
      Tabs[iCnt] := Tabs[iCnt + iCount];
    Dec(FTabCount, iCount);
  end;
  Result := TabCount;
end;

function TRuler.GetActualPos(iValue: Integer): Integer;
begin
  Result := iValue + dbPositive(CornerSize) - Offset;
end;

function TRuler.GetRulerPos(iValue: Integer): Integer;
begin
  Result := iValue - dbPositive(CornerSize) + Offset;
end;

function TRuler.GetTabByInch(iIndex: Integer): Real;
begin
  Result := -1;
  if (iIndex < 0) or (iIndex >= TabCount) then exit;
  if Tabs[iIndex] >= 0 then Result := Tabs[iIndex] / Screen.PixelsPerInch;
end;

function TRuler.GetTabByPixel(iIndex: Integer): Integer;
begin
  Result := -1;
  if (iIndex < 0) or (iIndex >= TabCount) then exit;
  if Tabs[iIndex] >= 0 then Result := Tabs[iIndex];
end;

function TRuler.GetTabBySegment(iIndex: Integer): Integer;
begin
  Result := -1;
  if (iIndex < 0) or (iIndex >= TabCount) then exit;
  if Tabs[iIndex] >= 0 then
    Result := Tabs[iIndex] div (Screen.PixelsPerInch div SegmentCount);
end;

function TRuler.PointOnTab(iX, iY: Integer): Integer;
var
  iCnt, iHalf, iLeft, iTop: Integer;
  aRect: TRect;
begin
  Result := -1;
  if (TabsEnabled = False) or PtInRect(FCornerRect, Point(iX, iY)) then exit;
  if Direction = dirHorz then begin
    if TabPosition = tpTop then iTop := 1
    else iTop := Height - TabGlyph.Height - 1;
    iHalf := TabGlyph.Width div 2;
  end else begin
    if TabPosition = tpLeft then iLeft := 1
    else iLeft := Width - TabGlyph.Width - 1;
    iHalf := TabGlyph.Height div 2;
  end;
  for iCnt := 0 to FTabCount-1 do begin
    if Direction = dirHorz then
      iLeft := Tabs[iCnt]+ dbPositive(CornerSize) - iHalf - Offset
    else
      iTop := Tabs[iCnt] + dbPositive(CornerSize) - iHalf - Offset;
    aRect := Rect(iLeft, iTop, iLeft + TabGlyph.Width, iTop + TabGlyph.Height);
    if PtInRect(aRect, Point(iX, iY)) then begin
      Result := iCnt;
      exit;
    end;
  end;
end;

procedure TRuler.ResetTabs;
var
  iCnt: Integer;
begin
  for iCnt := 0 to MAXTABS-1 do FTabs[iCnt] := -1;
  FTabCount := 0;
  Invalidate;
end;

function TRuler.SetTabsByInch(iFirst, iCount: Integer;
 var rValues: TRealArray): Integer;
var
  iCnt, iValue: Integer;
begin
  if iFirst + iCount > MAXTABS then iCount := MAXTABS - iFirst;
  Result := iCount;
  FTabCount := iCount;
  for iCnt := iFirst to iFirst + iCount do begin
    iValue := Trunc(Screen.PixelsPerInch * rValues[iCnt - iFirst]);
    FTabs[iCnt] := iValue;
  end;
  Invalidate;
end;

function TRuler.SetTabsByPixel(iFirst, iCount: Integer;
 var iValues: TIntArray): Integer;
var
  iCnt: Integer;
begin
  if iFirst + iCount > MAXTABS then iCount := MAXTABS - iFirst;
  Result := iCount;
  FTabCount := iCount;
  for iCnt := iFirst to iFirst + iCount do
    FTabs[iCnt] := iValues[iCnt - iFirst];
  Invalidate;
end;

function TRuler.SetTabsBySegment(iFirst, iCount: Integer;
 var iValues: TIntArray): Integer;
var
  iCnt, iValue: Integer;
begin
  if iFirst + iCount > MAXTABS then iCount := MAXTABS - iFirst;
  Result := iCount;
  FTabCount := iCount;
  for iCnt := iFirst to iFirst + iCount do begin
    iValue := (Screen.PixelsPerInch div SegmentCount) * iValues[iCnt - iFirst];
    FTabs[iCnt] := iValue;
  end;
  Invalidate;
end;



procedure Register;
begin
  RegisterComponents('DinoByte', [TRuler]);
end;

end.
